/*
 * file StpProvider.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM 
 *
 * com.ibm.rational.wvcm.stp.StpProvider
 *
 * (C) Copyright IBM Corporation 2004, 2011.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.rational.wvcm.stp;

import java.io.File;
import java.util.Date;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;

import javax.wvcm.PropertyNameList.PropertyName;
import javax.wvcm.ProviderFactory.Callback;
import javax.wvcm.ProviderFactory.Callback.Authentication;
import javax.wvcm.Resource;
import javax.wvcm.ResourceList;
import javax.wvcm.WvcmException;

import com.ibm.rational.wvcm.stp.StpAccessControlEntry.AccessRight;
import com.ibm.rational.wvcm.stp.StpException.StpReasonCode;
import com.ibm.rational.wvcm.stp.cc.CcProvider;
import com.ibm.rational.wvcm.stp.cq.CqProvider;
import com.ibm.rational.wvcm.stpex.StpExEnumeration;
import com.ibm.rational.wvcm.stpex.annotation.RpcEnum;


/**
 * StpProvider is an extension to the WVCM {@link javax.wvcm.Provider Provider}
 * interface. It defines additional features supported by providers of this API
 * that are common to both the ClearCase and ClearQuest domains. The StpProvider
 * implementation <i>classes</i> that are available for instantiation are named
 * by the String fields of this interface and its extensions and must be
 * instantiated using the WVCM
 * {@link javax.wvcm.ProviderFactory ProviderFactory} class.
 * 
 * <p>
 * For more flexible deployment of the API implementation, this library is
 * partitioned into two domain-specific subcomponents or <i>subproviders</i>.
 * There is at least one subprovider for each {@link StpProvider.Domain}. If
 * access to a given domain is available along multiple paths--e.g., locally or
 * via WAN--there will likely be one subprovider for each different access
 * method. The services of a single subprovider interface will be available only
 * if its subprovider class can be instantiated and such a subprovider can be
 * instantiated only if the Rational product underlying the subprovider's domain
 * has been properly installed and licensed. Each subprovider is autonomous,
 * however, and this provider can be instantiated as long as at least one of its
 * subproviders can be instantiated. If a client needs to know which domains are
 * supported by an instance of this provider, it may use the
 * {@link #isSupported(com.ibm.rational.wvcm.stp.StpProvider.Domain) isSupported}
 * method to discover this and use {@link #getInstantiationErrors()} for more
 * information why a subprovider failed to instantiated.
 * </p>
 * 
 * <p>
 * {@link javax.wvcm.ProviderFactory#createProvider(String,
 * javax.wvcm.ProviderFactory.Callback) ProviderFactory.createProvider} may be
 * invoked with a non-<b>null</b> Callback object from which an {@link
 * javax.wvcm.ProviderFactory.Callback.Authentication Authentication} object can
 * be obtained for each authentication realm visited by the provider. The
 * <b>realmId</b> provided to the
 * {@link javax.wvcm.ProviderFactory.Callback#getAuthentication getAuthentication}
 * method tells the user what credentials to provide. For ClearQuest, the realm
 * identifier is the database set name (also known as the schema repository,
 * master database, profile, or connection name); for ClearCase, the realm
 * identifier is the URL for the server.
 * </p>
 * <p>
 * This interface defines {@link StpCallback}, which extends
 * {@link javax.wvcm.ProviderFactory.Callback} to allow the library to provide
 * more information to the client's callback object when it is asked for
 * credentials.
 * </p>
 * If no Callback instance is made available to the provider, operations that
 * require an authenticated identity will fail unless the client has already
 * provided an Authentication object to the provider via the
 * {@link #setAuthentication(com.ibm.rational.wvcm.stp.StpProvider.Domain, String, javax.wvcm.ProviderFactory.Callback.Authentication, boolean) setAuthentication}
 * method
 * <p>
 * Some subproviders require the client to tell them the URL for their
 * associated server. That is accomplished using the
 * {@link #setServerUrl(String) setServerUrl} method following Provider
 * construction. Setting a server URL does not result in any validation that the
 * server is accessible. A client must perform some operation (e.g.,
 * {@link javax.wvcm.Resource#doReadProperties(javax.wvcm.Feedback)}) to
 * attempt to use, and hence check, a server URL. An attempt to create a
 * Location associated with such a subprovider will fail if the server URL has
 * not yet been specified. The server URL associated with a subprovider cannot
 * be changed once it has been set via {@link #setServerUrl}. A new StpProvider
 * instance must be created to reference a different server.
 * </p>
 * 
 * <p>
 * When a Provider is instantiated via the ProviderFactory, a new, independent
 * sequence of server interactions is started for a single user. The user of the
 * provider is identified by the {@link javax.wvcm.ProviderFactory.Callback
 * Callback} instance passed through the ProviderFactory to the Provider
 * instance.
 * </p>
 * 
 * <p>
 * A provider exists until it is explicitly terminated via the {@link
 * #terminate() terminate()} method or it is implicitly terminated after a
 * period of time in which no server interactions are initiated. It is always 
 * best to explicitly terminate a provider when it is no longer needed. This
 * will free up resources on the server for others to use.
 * </p>
 */
public interface StpProvider extends javax.wvcm.Provider
{
    /**
     * The name of a wvcm.javax.Provider implementation class to be used by
     * clients that wish to interact with both ClearCase and ClearQuest in a
     * single provider. The returned object supports only the StpProvider
     * interface and therefore does not provide direct access to any
     * domain-specific capabilities. Once this generic provider has been
     * instantiated, the client can obtain a domain-specific provider extension
     * using either the {@link #cqProvider()} method or the
     * {@link #ccProvider()} method.
     * <p>
     * A client could create this provider to discover which team services are
     * installed. The domain-specific provider access methods of this instance
     * will return <b>null</b> if the domain-specific provider is not
     * installed.
     * 
     * @see com.ibm.rational.wvcm.stp.cq.CqProvider#CQ_ONLY_PROVIDER_CLASS
     * @see com.ibm.rational.wvcm.stp.cc.CcProvider#LOCAL_PROVIDER_CLASS
     * @see com.ibm.rational.wvcm.stp.cc.CcProvider#NETWORK_PROVIDER_CLASS
     */
    String PROVIDER_CLASS =
        "com.ibm.rational.stp.client.internal.core.StpProviderImpl";

    /**
     * @return The Provider extension associated with this StpProvider that 
     * implements the CcProvider interface; <b>null</b> if that Provider 
     * extension is not instantiated with this provider.
     */
    CcProvider ccProvider();

    /**
     * @return The Provider extension associated with this provider instance
     *         that implements the CqProvider interface; <b>null</b> if that
     *         Provider extension is not instantiated with this provider.
     */
    CqProvider cqProvider();

    /**
     * @return The Provider extension associated with this Provider instance
     *         that implements the StpProvider interface; will never be <b>null</b>.
     */
    StpProvider stpProvider();

    /**
     * Version information about the code that implements a portion of this
     * product. This information is available for this Provider, and each of
     * it's domain-specific subproviders. Version information for the server
     * used to access a given repository is available as the PRODUCT_INFO
     * property of those repositories.
     * <p>
     * Any field not applicable to the code for which information is requested
     * will be an empty String (not <b>null</b>).
     */
    static public interface StpProductInfo
    {
        /**
         * @return The product build number, representing the finest granularity
         *         of the product version specification.
         */
        String getBuildNumber();

        /**
         * @return the e-mail address of the company for the current locale.
         */
        String getCompanyEmailAddress();

        /**
         * @return the full name of the company in the current locale.
         */
        String getCompanyFullName();

        /**
         * @return the name of the company in the current locale.
         */
        String getCompanyName();

        /**
         * @return the web address of the company for the current locale.
         */
        String getCompanyWebAddress();

        /**
         * @return the full product version identifier string, containing both
         * the product version and the build number.
         */
        String getFullProductVersion();

        /**
         * @return the current fix pack version of the product.
         */
        String getPatchVersion();

        /**
         * @return the version identifier for the code, without build number.
         */
        String getProductVersion();

        /**
         * @return the ClearCase label used to uniquely identify each build.
         */
        String getStageLabel();
    }

    /**
     * Returns the version information for the code implementing this Provider
     * or one of its domain-specific subproviders. Each subprovider may make
     * available additional version information as extensions to the
     * StpProductInfo interface.
     * 
     * @param domain The domain for which subprovider information is desired. If
     *            <b>null</b>, the version information for this interface and
     *            the shared components of this provider will be returned.
     * 
     * @return An StpProductInfo structure (or a domain-specific extension
     *         thereof) containing version information for the code that
     *         implements this Provider (domain == <b>null</b>) or for the code
     *         that implements the subprovider for the domain specified by the
     *         domain argument. Will be <b>null</b> if a subprovider for the
     *         specified domain has not be instantiated.
     */
    StpProductInfo stpProductInfo(Domain domain);

    /**
     * Constructs an StpRepository proxy for a repository at a specified
     * location.
     * 
     * @param location The location the new proxy is to refer to. A repository
     *            need not exist at this location, but the location must be
     *            otherwise valid for a repository. Must not be <b>null</b>.
     * 
     * @return A new StpRepository proxy for a repository at the specified
     *         location.
     */
    StpRepository stpRepository(StpLocation location);

    /**
     * Returns the Class object for the value type of a property identified by a
     * PropertyName. Only properties specified explicitly by this API are
     * supported. For all other properties, the {@link StpProperty#TYPE TYPE}
     * meta-property may be useful.
     * 
     * @param name The PropertyName for the property whose value class is being
     *            requested.
     * 
     * @return a Class object identifying the value type of the property.
     *         Will be <b>null</b> if the PropertyName or value type is unknown.
     */
    <T> Class<T> getPropertyValueClass(PropertyName<T> name);

    /**
     * Constructs a new proxy of a given type for a resource at a specified
     * location, defining its {@link javax.wvcm.Resource#RESOURCE_IDENTIFIER}
     * property value if known.
     * 
     * <p>
     * This method and StpResource.initMetaProperty provide the means for
     * reconstructing a proxy previously archived to persistent storage by a
     * client. When the proxy is archived, the client needs to save its
     * location, its proxyType, the value of its RESOURCE_IDENTIFIER property
     * (if defined), and the name and value of any other property/meta-property
     * defined by the proxy that will be of interest to the client when the
     * proxy is restored.
     * </p>
     * 
     * <p>
     * To restore the proxy, invoke this method with the previously saved
     * location, proxyType and RESOURCE_IDENTIFIER value to get an appropriate
     * proxy object. Then use StpResource.initMetaProperty to move all of the
     * remaining saved properties into the proxy.
     * </p>
     * 
     * @param location The resource location to which the new proxy is to refer.
     *            Must not be <b>null</b>. A resource need not exist at the
     *            specified location, but the location must otherwise be valid.
     * @param proxyType A String specifying the type of proxy required. Must not
     *            be <b>null</b>. This value may be obtained either from a
     *            previous proxy for the resource using its {#link
     *            StpResource#proxyType()} method or from the Class object for
     *            the public interface that the new proxy should support using
     *            the {@link #proxyType(Class)} method.
     * @param resourceIdentifier The value of the RESOURCE_IDENTIFIER property
     *            to be defined by the new proxy. This value would have to come
     *            from the RESOURCE_IDENTIFIER value of a pre-existing proxy. If
     *            <b>null</b>, the RESOURCE_IDENTIFIER property of the new
     *            proxy will not be defined. This property cannot be set after
     *            the proxy is constructed.
     * 
     * @return A new proxy implementing the interfaces implied by the given
     *         proxyType parameter. Will not be <b>null</b>.
     * 
     * @throws WvcmException Thrown if an StpLocation object cannot be obtained
     *             from this provider for the given location or if this provider
     *             cannot construct a proxy that implements the interface
     *             implied by the proxyType argument.
     * 
     * @see StpResource#initMetaProperty(javax.wvcm.PropertyNameList.PropertyName,
     *      com.ibm.rational.wvcm.stp.StpProperty.MetaPropertyName, Object)
     */
    StpResource buildProxy(StpLocation location,
                           String proxyType,
                           String resourceIdentifier) throws WvcmException;

    
    /**
     * Constructs a proxy implementing a given Resource interface class, using a
     * String representation of the location.
     * 
     * @param <T> The Resource interface type that the returned proxy is to
     *            implement. This is inferred from the type argument.
     * @param type The Class object for the interface class that the returned
     *            proxy is to implement.
     * @param str A String containing a specification for the location of the
     *            resource to be addressed by the returned proxy. The domain and
     *            namespace do not need to be specified in this location since
     *            they can be inferred from the desired proxy class. The
     *            repository field may also be elided if a default repository
     *            has been established for this provider.
     * @return A proxy object that implements interface T.
     * @throws WvcmException Thrown if an StpLocation object cannot be obtained
     *             from this provider for the given location or if this provider
     *             cannot construct a proxy that implements interface T.
     * 
     * 
     * @see #setDefaultRepository(com.ibm.rational.wvcm.stp.StpProvider.Domain, StpRepository)
     * @see javax.wvcm.Provider#buildProxy(Class, javax.wvcm.Location)
     */
    <T extends Resource> T buildProxy(Class<T> type,
                                      String str) throws WvcmException;

    /**
     * Computes the proxy type string required by
     * {@link #buildProxy(StpLocation, String, String)} given the Class object
     * for a proxy interface that the new proxy should implement.
     * 
     * <p>
     * Note that
     * 
     * <pre>
     * provider.buildProxy(location,
     *                     provider.proxyType(Xyz.class),
     *                     null)
     * </pre>
     * 
     * is equivalent to
     * 
     * <pre>
     * provider.xyz(location)
     * </pre>
     * 
     * where <code>provider.xyz()</code> is the proxy factory method that
     * returns an implementation of the interface Xyz. This facilitates the
     * building of proxies driven by compile-time data structures by eliminating
     * the need for the client to develop program logic to switch on such data
     * down to an appropriate proxy factory method call.
     * </p>
     * 
     * @param interfaceClass A public proxy class (i.e. any Class object for
     *            which
     *            javax.wvcm.Resource.class.isAssignableFrom(interfaceClass) is
     *            <b>true</b>).
     * 
     * @return A String suitable for use as the proxyType argument to
     *         {@link #buildProxy(StpLocation, String, String)}.
     * 
     * @throws WvcmException If the given Class is not an interface implemented
     *             by a proxy available from this provider.
     *             
     * @see #buildProxy(Class, String)
     * @see #buildProxy(StpLocation, String, String)
     */
    String proxyType(Class<? extends Resource> interfaceClass)
        throws WvcmException;

    /**
     * The domain to use if the domain field is omitted from location
     * specifications.
     * 
     * @return The StpProvider.Domain enumerator representing the domain that
     *         will be used if none is specified in a location specification.
     */
    StpProvider.Domain getDefaultDomain();

    /**
     * Establishes the default StpProvider.Domain for this provider. The default
     * domain is applied when a location missing a domain is used to construct a
     * proxy.
     * 
     * @param domain The default StpProvider.Domain. Use <b>null</b> to disable
     *            the automatic filling-in of missing domain fields.
     * 
     * @throws WvcmException if this provider does not support the given domain.
     */
    void setDefaultDomain(Domain domain) throws WvcmException;

    /**
     * Returns the default repository associated with this provider.
     * 
     * @return A proxy for the default repository established for this provider.
     *         Will be <b>null</b> if no default has been established.
     */
    StpRepository getDefaultRepository();

    /**
     * Establishes the default repository to be associated with this provider.
     * The default repository is applied when a location missing a repository
     * field is used to construct a proxy.
     * 
     * @param repo The new default repository. May be <b>null</b> to indicate
     *            that there is no default repository associated with this
     *            provider.
     * 
     * @throws WvcmException if this provider does not support the repository
     *             type of the given repository.
     */
    void setDefaultRepository(StpRepository repo) throws WvcmException;

    /**
     * Returns the default repository of a specified domain associated with this
     * provider.
     * 
     * @param domain The StpProvider.Domain for which a default repository is
     *            being requested.
     * 
     * @return A proxy for the default repository established for this provider.
     *         Will be <b>null</b> if no default has been established.
     */
    StpRepository getDefaultRepository(StpProvider.Domain domain);

    /**
     * Establishes the default repository of a specified type to be associated
     * with this provider.
     * 
     * @param domain The StpProvider.Domain for which the given repository is to
     *            be the default.
     * @param repo The new default repository. May be <b>null</b> to
     *            indicate that there is no default repository associated with
     *            this provider.
     * 
     * @throws WvcmException if this provider does not support the given
     *             domain.
     */
    void setDefaultRepository(StpProvider.Domain domain,
                              StpRepository repo) throws WvcmException;

    /**
     * The key for the initial argument map entry that contains the lowercase
     * two-letter ISO-639 code for the language of the client locale for this
     * provider.
     * 
     * @see #setUserLocale(Locale)
     * @see java.util.Locale
     */
    String USER_LANGUAGE_KEY = "user-language";

    /**
     * The key for the initial argument map entry that contains the uppercase
     * two-letter ISO-3166 code for the country of the client locale for this
     * provider.
     * 
     * @see #setUserLocale(Locale)
     * @see java.util.Locale
     */
    String USER_COUNTRY_KEY = "user-country";

    /**
     * The key for the initial argument map entry that contains the variant
     * vendor and browser specific code for the client locale of this provider.
     * 
     * @see #setUserLocale(Locale)
     * @see java.util.Locale
     */
    String USER_VARIANT_KEY = "user-variant";

    /**
     * The key for the initial argument map entry that contains the server URL
     * for this provider.
     * 
     * @see #setServerUrl(String)
     */
    String SERVER_URL_KEY = "server-url";

    /**
     * The key for the initial argument map entry that indicates whether or not
     * the provider is to operate disconnected from a server.
     */
    String IS_DISCONNECTED_KEY = "is-disconnected";

    /**
     * The value for the initial argument map entry that indicates 
     * the provider is not allowed to connected to a server.
     */
    String IS_DISCONNECTED_VALUE = "true";

    /**
     * The value for the initial argument map entry that indicates 
     * the provider is allowed to connect to a server.
     */
    String NOT_DISCONNECTED_VALUE = "false";

    /**
     * Establishes the default server URL for this provider, if one has not
     * already been established. Within the lifetime of a given instance of an
     * StpProvider, the server URL may be set only once. This method will throw
     * an exception if the provider's server URL has already been set when
     * applied.
     * <p>
     * The provider's server URL may also be applied as an initialization
     * argument using the key {@link #SERVER_URL_KEY}.
     * 
     * @param url The URL used to contact the CM Server via this provider. Must
     *            not be <b>null</b>. This URL is not validated until the next
     *            "do" method that requires a URL is invoked.
     * 
     * @throws WvcmException if a server URL has already been established for
     *             this provider.
     */
    void setServerUrl(String url) throws WvcmException;

    /**
     * Returns the server URL for this provider, as set by an initialization
     * argument or a previous call to setServerUrl(). If no default server URL
     * has been established, returns <b>null</b>.
     * 
     * @return the default server URL for this provider; may be <b>null</b>.
     * 
     * @see #setServerUrl(String)
     */
    String getServerUrl();

    /**
     * Returns the Locale that is being used by this provider and is being
     * passed to the server to identify the locale to use when performing
     * locale-dependent operations for this provider. If not explicitly set by
     * the client, the value returned by java.util.Locale#getDefault() is
     * returned.
     * 
     * @return A Locale object representing the locale to be used when
     *         performing locale-dependent operations for this provider. Will
     *         not be <b>null</b>.
     */
    Locale getUserLocale();

    /**
     * Defines the Locale that is to be used by this provider and passed to the
     * server to identify the locale to use when performing locale-dependent
     * operations for this provider. Does <i>not</i> set the default Locale of
     * the JVM.
     * 
     * @param locale A Locale object representing the locale to be used when
     *            performing locale-dependent operations for this provider. Must
     *            not be <b>null</b>; use java.util.Locale#getDefault() to use
     *            the native Locale of the client.
     */
    void setUserLocale(Locale locale);

    /**
     * Returns the TimeZone that is being used by this provider and is being
     * passed to the server to identify the time zone of the user. If not
     * explicitly set by the client, java.util.TimeZone#getDefault() is
     * returned.
     * 
     * @return A TimeZone object representing the time zone to be used when
     *         performing date translations for this provider. Will not be
     *         <b>null</b>.
     */
    TimeZone getUserTimeZone();

    /**
     * Defines the time zone that is to be used by this provider and passed to
     * the server for this provider. Does NOT set the default TimeZone of the
     * JVM.
     * 
     * @param timeZone A TimeZone object representing the time zone to be used
     *            when performing date translation for this provider. Must not
     *            be <b>null</b>; use java.util.TimeZone#getDefault() to use
     *            the native time zone of the client.
     */
    void setUserTimeZone(TimeZone timeZone);

    /**
     * The type of a subprovider.
     */
    static public enum Domain implements StpExEnumeration
    {
        /** An unspecified domain */
        @RpcEnum(id=3, simpleName="")
        NONE(""),

        /** An invalid domain */
        @RpcEnum(id=2, simpleName="<invalid>")
        INVALID("<invalid>"),

        /** The ClearCase domain. */
        @RpcEnum(id=0, simpleName="cc")
        CLEAR_CASE("cc"),

        /** The ClearQuest domain */
        @RpcEnum(id=1, simpleName="cq")
        CLEAR_QUEST("cq");
        
        /**
         * Returns the Domain object that matches a given domain-field symbol
         * 
         * @param symbol A String containing the domain-field symbol to be
         *            converted to a Domain object.
         * 
         * @return A Domain object that matches the given symbol. Will be <b>
         *         INVALID</b> if no Domain object matches the given symbol.
         */
        public static final Domain fromSymbol(String symbol)
        {
            if (symbol != null) {
                for (Domain d : values())
                    if (symbol.compareToIgnoreCase(d.toSymbol()) == 0)
                        return d;
            }

            return INVALID;
        }

        /**
         * @return Returns the word token to use in the domain field of a
         *         location specification for this Domain.
         */
        public String toSymbol()
        {
            return m_symbol;
        }

        /**
         * Creates a new Domain object.
         * 
         * @param symbol The symbol used for this StpProvider.Domain when
         *            forming a location specification.
         */
        private Domain(String symbol)
        {
            m_symbol = symbol;
        }

        /* The domain-field symbol that corresponds to the Domain enumerator */
        private String m_symbol;
    }

    /**
     * Returns an indication of whether or not this provider is permitted to
     * communicate with a server.
     * 
     * @return <b>true</b> if this provider is permitted to connect to a
     *         server to perform operations; <b>false</b> otherwise.
     */
    boolean getIsDisconnected();

    /**
     * Grants or denies this provider the ability to connect to a server while
     * performing tasks for the client.
     * <p>
     * Note: Setting this attribute does <i>not</i> cause an interaction with
     * the server to initialize or terminate the session. Use the
     * {@link #setAuthentication(com.ibm.rational.wvcm.stp.StpProvider.Domain, String, javax.wvcm.ProviderFactory.Callback.Authentication, boolean) setAuthentication}
     * method to control session lifetimes.
     * 
     * @param isDisconnected <b>true</b> if this provider is not to connect to
     *            server; <b>false</b> if this provider may connect to a
     *            server.
     */
    void setIsDisconnected(boolean isDisconnected);

    /**
     * Returns the callback object passed to the ProviderFactory during the
     * instantiation or initialization of this provider.
     * 
     * <p>
     * The provider maintains a mapping between authentication realms and user
     * credentials and this callback object is used to populate that map. When
     * the library generates a request targeting a specific Location, it will
     * attempt to deduce the realm id from the Location (through its
     * subprovider).
     * </p>
     * 
     * <ol>
     * <li>If a realm id can be deduced, the provider will consult the
     * realm-credential map to obtain the credentials for that realm id.
     * <ul>
     * <li>If no credentials are found for the realm, this authentication
     * callback will be called passing the realm id to the client. The returned
     * credentials will then be entered into the realm-credential map and the
     * request will proceed with those credentials.</li>
     * <li>If credentials are found for the realm, the request proceeds with
     * those credentials.</li>
     * </ul>
     * </li>
     * <li>If the realm id cannot be deduced from the Location, the value of
     * the AUTHENTICATION_REALM property will be examined.
     * <ul>
     * <li>If the AUTHENTICATION_REALM property is defined, the request will be
     * submitted using its value as the realm id as outlined in bullet 1.</li>
     * <li>If the AUTHENTICATION_REALM property is not defined, the library
     * issues the request without credentials.
     * 
     * <ul>
     * <li>If the request succeeds, the credentials used are mapped to the
     * value returned in the AUTHENTICATION_REALM property (if the response
     * contains this property).</li>
     * <li>If the request fails, the request will be resubmitted using the
     * realm id provided in the error response as outlined in bullet 1.</li>
     * </ul>
     * </li>
     * </ul>
     * </li>
     * <li>If a request fails authentication when using credentials from the
     * realm-credential map, this authentication callback is called again for a
     * new set of credentials for the realm.
     * <ul>
     * <li>If a new set of credentials is not returned, the request will fail
     * with an authentication exception.</li>
     * <li>If new credentials are returned, the realm-credential map is updated
     * with the new credentials and the request is retried as outlined in bullet
     * 1.</li>
     * </ul>
     * </li>
     * </ol>
     * 
     * <p>
     * Rather than embedding the policy for establishing the default credentials
     * in the library, the library always uses the authentication callback (with
     * a <b>null</b> realm) to obtain from the client the <i>default credentials</i>
     * to use. The client can then implement the strategy it wishes to use to
     * come up with the default credentials.
     * </p>
     * 
     * @return The call-back object used to initiate the provider.
     */
    javax.wvcm.ProviderFactory.Callback callback();

    /**
     * Returns information about the failures, if any, to instantiate providers
     * for specific product domains. If the map returned contains no information
     * (i.e. <b>null</b>) for a given StpProvider.Domain, it is safe to use the
     * type-specific interfaces for that domain.
     * <p>
     * Some domains may be served by more than one type of provider--ClearQuest
     * repositories, for example, can be accessed via the network using one
     * provider or accessed via an installed ClearQuest application using
     * another. The StpProvider.Domain entry in the returned error map will be
     * <b>null</b> if at least one subprovider for the type could be
     * instantiated. It will be non-<b>null</b> only if <i>all</i>
     * subproviders for that domain failed to instantiate.
     * <p>
     * For more thorough diagnosis of instantiation problems, an entry will be
     * found in the map associated with the toSymbol() String value of a
     * StpProvider.Domain if <i>any</i> provider for that type failed to
     * instantiate.
     * 
     * @return A mapping from StpProvider.Domain objects and String objects to
     *         an StpException object containing pertinent information about the
     *         failed attempt to instantiate the provider for the given type of
     *         repository. Will never be <b>null</b>, but may be empty or
     *         sparse.
     */
    Map<Object, StpException> getInstantiationErrors();

    /**
     * Terminates this provider's association with the servers that it is
     * actively tracking, releasing any resources it has accumulated as a result
     * of that tracking.
     * 
     * <p>
     * Calling this method marks all of the provider's proxy and location
     * objects as invalid, and so all client-side references should be
     * abandoned; any subsequent method calls on the Provider object yields an
     * UnsupportedOperationException exception. Thus you should typically call
     * this method immediately prior to the exit of the owning thread or
     * process.
     * </p>
     * 
     * @throws WvcmException if the accumulated resources cannot be disposed of.
     */
    void terminate() throws WvcmException;

    /**
     * Updates the realm-authentication map maintained by this Provider by
     * adding or deleting associations between realm strings and Authentication
     * objects from which user credentials for that realm can be obtained when
     * they are needed by an operation.
     * <p>
     * If a Callback object has been assigned to the Provider, the credential
     * processing proceeds as follows:
     * <p>
     * When credentials are needed to perform an operation in a specific realm,
     * the Provider first consults the realm-authentication map for an
     * Authentication object associated with the realm. If there is no
     * Authentication object registered in the map, the Provider calls the
     * {@link javax.wvcm.ProviderFactory.Callback#getAuthentication(String, int) Callback.getAuthentication()}
     * method with a retryCount of 0. If there is an Authentication object in
     * the map for the realm and the credentials it provides are rejected by the
     * server, the Authentication object is removed from the map and the
     * Provider calls Callback.getAuthentication() (with a retryCount > 1) to
     * obtain a new Authentication object to associate with the realm.
     * <p>
     * If no Callback object has been assigned to the Provider, the credential
     * processing proceeds as follows:
     * <p>
     * When credentials are needed to perform an operation in a specific realm,
     * the Provider consults the realm-authentication map for an Authentication
     * object associated with the realm. If there is no Authentication object
     * registered in the map no credentials will be passed to the server. If the
     * chosen Authentication object is rejected by the server, the Provider
     * throws an exception to terminate the operation.
     * <p>
     * The Provider adds an entry to the realm-authentication map only after it
     * has determined that a given Authentication object actually provides valid
     * credentials in the realm. It will remove an entry from the map when it
     * determines that the Authentication object in the map cannot provide valid
     * credentials (and there is a Callback object from which to get a
     * replacement).
     * <p>
     * The client may make entries in the realm-authentication map with or
     * without validating those entries. Validation implies a round-trip to the
     * server for the realm(s) specified. If the realm-authentication map
     * contains an entry for an affected realm, the corresponding server is
     * notified to log the user out of that realm. If the Authentication object
     * is not <b>null</b>, the server is notified to log the user into the
     * specified realm with the new set of credentials.
     * 
     * In this release, the realm string for a ClearQuest database is just the
     * database set name. It is assumed that all user databases defined in that
     * database set use the same credentials as stored in the database set. (In
     * fact, there are times in the history of a given user database when the
     * credentials it accepts for a user differ from those stored in its
     * database set. This corner case is not supported in this release.)
     * 
     * A full user database name may be passed to this method, but the
     * authentication map remembers only the Db-Set portion of it. The user
     * database portion of the name, if present, is used only when doValidation
     * is true. In that case, if the realm parameter includes a user database
     * name, then the user is logged into or out of that user database (not the
     * database set).
     * 
     * @param domain The Domain for which authentication is being provided. Must
     *            not be <b>null</b> unless the authentication parameter is
     *            also <b>null</b>. In that case, the Authentication object is
     *            removed for the given realm in all domains.
     * @param Realm A String identifying the realm for which authentication is
     *            being provided. (This is the string passed as the realm
     *            parameter to Callback.getAuthentiation()) Must not be <b>null</b>
     *            unless the authentication parameter is also <b>null</b>, in
     *            which case the authentications for all realms (in the
     *            specified domains) will be removed from the map.
     * @param authentication A ProviderFactory.Authentication instance that can
     *            provide any credentials needed for the user in the specified
     *            realm of the specified domain(s). If <b>null</b>, any
     *            existing authentication associated with the specified realm(s)
     *            will be removed from the map.
     * @param doValidation If <b>false</b>, only the Provider map will be
     *            updated; if <b>true</b>, each impacted domain server will be
     *            contacted immediately to inform it of the change in
     *            credentials. The user will be logged off the realm(s) already
     *            in the map and, if the Authentication object is not <b>null</b>
     *            a new logon will be attempted using the given credentials.
     * @throws WvcmException
     *             <ul>
     *             <li>if doValidation is <b>true</b>, the provided
     *             Authentication object is not <b>null</b>, and the
     *             credentials it provides are rejected by a server, or
     *             <li> if doValidation is <b>true</b>, the provided
     *             Authentication object is <b>null</b>, and a server cannot
     *             terminate the provider of (i.e. logoff) a once-active user.
     *             </ul>
     */
    void setAuthentication(Domain domain,
                           String Realm,
                           Authentication authentication,
                           boolean doValidation) throws WvcmException;

    /**
     * Returns the Authentication object currently associated with a given realm
     * string in a given Domain.
     * 
     * @param domain The Domain for which authentication is being requested;
     *            Must not be <b>null</b>;
     * @param realm The realm for which authentication is being requested; Must
     *            not be <b>null</b>.
     * @return The Authentication object currently associated with the given
     *         Domain and realm; <b>null</b> if no such association currently
     *         exists.
     */
    Authentication getAuthentication(Domain domain,
                                     String realm);

    /**
     * An extended version of javax.wvcm.ProviderFactory.Callback whose
     * getAuthentication method provides additional information to the client
     * when being asked for credentials. If the Callback object passed to this
     * Provider implements this interface, its getAuthenticationEx will be
     * called rather than Callback.getAuthentication each time an Authentication
     * object is required.
     */
    public interface StpCallback extends Callback
    {
        /**
         * An extended version of Callback.getAuthentication that provides
         * additional information to the client when being asked for
         * credentials. If the Callback object passed to this Provider
         * implements this interface, the getAuthenticationEx will be called
         * rather than getAuthentication each time an Authentication object is
         * required.
         * 
         * @param domain The Domain that is requesting credentials.
         * @param realm an identifier for the authentication realm that is
         *            requesting credentials. A single provider instance can
         *            access resources in different authentication realms, so
         *            the user must be provided with the name of the
         *            authentication realm in order to determine what
         *            credentials to provide.
         * @param retryCount the number of times the provider has unsuccessfully
         *            attempted to get authentication for this operation. This
         *            parameter helps the client decide when to provide a more
         *            detailed authentication dialog, or potentially abort the
         *            operation rather than making additional attempts to get
         *            credentials from the user.
         * @param provider The StpProvider instance that is requesting
         *            authentication.
         * @param failure The WvcmException generated by the attempt to log onto
         *            the realm using the Authentication object last provided by
         *            this Callback. This will be <b>null</b> when retryCount
         *            is zero.
         * @return user and password authentication interface. Must not be
         *         <b>null</b>
         * @throws WvcmException if the user or client wish to abort the
         *             authentication request. Note: <i>this method throwing an
         *             exception is the only way the authentication retry
         *             process can be terminated.</i> The provider has no
         *             preset retry limits or conditions and will repeatedly
         *             call this Callback object until a valid set of
         *             credentials is returned or this method throws an
         *             exception.
         */
        public Authentication getAuthenticationEx(Domain domain,
                                                  String realm,
                                                  int retryCount,
                                                  StpProvider provider,
                                                  WvcmException failure)
            throws WvcmException;
    }

    /**
     * If the Callback object provided to an StpProvider implements this
     * interface, the callback will be notified when an Authentication object
     * returned by the Callback's getAuthentication() method is successfully
     * verified to be valid.
     */
    public interface NotifyAuthenticatedOption
    {
        /**
         * Notifies the Callback object that an authentication object that it
         * previously returned has been found to be valid for the specified
         * domain, realm, and provider. If the provided credentials were not
         * valid, the Callback's getAuthentication method will be called another
         * time for a new set of credentials.
         * 
         * @param domain The Domain that had requested credentials.
         * @param realm an identifier for the authentication realm that had
         *            requested credentials.
         * @param provider The StpProvider instance that had requested
         *            authentication.
         * @param authentication The credentials that were returned by the
         *            Callback and were accepted by the provider for the given
         *            domain and realm.
         */
        void notifyAuthenticated(Domain domain,
                                 String realm,
                                 StpProvider provider,
                                 Authentication authentication);
    }
    
    /**
     * If the Callback object provided to an StpProvider implements this
     * interface, the callback will be notified when the server reports that
     * it is too busy to respond to a request.
     */
    public interface NotifyBusyOption
    {
        /**
         * A non-<b>null</b> busyFailure argument indicates the server has
         * reported that it is too busy to process a request. This method
         * returns if and when that request is to be attempted again. It throws
         * an exception to cancel the request.
         * <p>
         * A <b>null</b> busyFailure argument indicates that a request that was
         * re-attempted has now been processed by the server. In this case, this
         * method must return and should not throw an exception if it wants the
         * operation to continue. This call with a <b>null</b> busyFailure
         * argument is intended to give the callback an opportunity to take down
         * any user presentation put up to indicate the operation was waiting
         * for the server to respond.
         * <p>
         * This method is <i>not</i> called for every server request, even with
         * a <b>null</b> busyFailure argument. It is invoked only when the
         * server reports that it is too busy to process a request.
         * <p>
         * The method returns no value. If the client wants to retry the request
         * a little later, the callback should put the current thread to sleep
         * for a while to give the server time to catch up. Whenever this method
         * returns, the request will be retried immediately.
         * <p>
         * If the client does not want to retry the request, it must throw an
         * exception, such as the busyFailure exception passed into the method.
         * 
         * @param domain
         *            The Domain in which the request was being executed.
         * @param realm
         *            The authentication realm in which the request was being
         *            executed.
         * @param provider
         *            The provider that is attempting to perform the request on
         *            the server.
         * @param opStartTime
         *            A Date object representing the time at which the request
         *            was first initiated. This object is unique to the request
         *            being executed and the same object is passed to every
         *            invocation of this method resulting from any attempt to
         *            re-exeucte the request. Thus, it could be used by the
         *            callback as the key to a map containing progress history
         *            for this operation. It could also be compared to the
         *            current time to limit the time the client waits for the
         *            server.
         * @param busyFailure
         *            The exception returned by the server to report that it was
         *            too busy to process the request. If this method returns
         *            normally and on the subsequent re-try the server processes
         *            the request this method is called with <b>null</b> for
         *            this argument. If the server still reports that it is busy
         *            this method will be called again with a non-null argument.
         *            <p>
         *            A <b>null</b> value for this argument <i>does not</i>
         *            necessarily mean that the operation was successful. It
         *            just means that the server was finally able to process the
         *            request. For example, after the server looks at the
         *            request, it may determine that authentication is required
         *            before proceeding or it may find something else wrong with
         *            it. This callback will be invoked to indicate that the
         *            busy state has been removed <i>before</i> processing the
         *            server response, whether it be positive or negative.
         *            <p>
         *            If completion of the operation requires multiple
         *            interactions with the server (as when authentication is
         *            needed), any of those interactions could find the server
         *            busy again. Hence, a <b>null</b> busyFailure argument
         *            <i>does not</i> indicate that this method will not be
         *            called another time for the same request (opStartTime).
         * @throws WvcmException to abort the request.
         */
        void notifyBusy(Domain domain,
                        String realm,
                        StpProvider provider,
                        Date opStartTime,
                        WvcmException busyFailure) throws WvcmException;
    }
    
    /**
     * Set the HTTP proxy host and port for all future calls from this
     * client.
     * @param host Name of proxy host.
     * @param port Port number of proxy on that host.
     */
    void setHttpProxy(String host,
                      String port);

    /**
     * @return A String containing the name of the HTTP proxy host used by this
     *         provider, <b>null</b> if not previously set.
     */
    String getHttpProxyHost();

    /**
     * @return A String containing the port number on the HTTP proxy host used
     *         by this provider, <b>null</b> if not previously set.
     */
    String getHttpProxyPort();

    /**
     * Determines whether a given domain is supported by this provider. It is
     * safe to use the type-specific interfaces for a supported domain.
     * 
     * @param domain - the domain in question
     * 
     * @return <b>true</b> if the domain is supported; <b>false</b> otherwise.
     * 
     * @see #getInstantiationErrors()
     * @see #ccProvider()
     * @see #cqProvider()
     */
    boolean isSupported(StpProvider.Domain domain);

    /**
     * Constructs an StpLocation from its string representation
     * 
     * @param location A string satisfying the syntax for a CM API location
     *            specification.
     * 
     * @return An StpLocation object capturing the fields specified in the given
     *         location specification string.
     * 
     * @throws WvcmException If the string is <b>null</b> or empty.
     */
    StpLocation stpLocation(String location) throws WvcmException;

    /**
     * Constructs an StpLocation from the individual fields of a location
     * specification that follows the user-friendly object selector scheme
     * 
     * @param domain the Domain. Must not be INVALID
     * @param namespace the Namespace for the location. Must not be INVALID,
     *            REPO, or FAST
     * @param name the segmented object name field. Each segment of the name
     *            must be encoded as required by the namespace. May be <b>null</b>
     *            or empty to form the root of the namespace in a repository
     *            resource.
     * @param repo the repository name. Each segment of the name must be encoded
     *            as required by the namespace.
     * 
     * @return An StpLocation object with the specified fields.
     * 
     * @throws WvcmException Thrown if the location String composed from the
     *             given fields is not in the correct form. StpReasonCode =
     *             {@link StpException.StpReasonCode#INVALID_OBJECT_SELECTOR}
     */
    StpLocation userFriendlySelector(StpProvider.Domain domain,
                                     StpLocation.Namespace namespace,
                                     String name,
                                     String repo) throws WvcmException;

    /**
     * Constructs an StpLocation from a string representation assuming the
     * Location is to be used exclusively in a specified domain. In particular,
     * if there is no scheme prefix or the given scheme prefix would not be 
     * recognized in the given domain, the input is forced to be a pname (in the
     * PNAME_IMPLIED namespace).
     * 
     * @param domain The domain in which the location is to be used. Must not be
     *            <b>null</b>, NONE, or INVALID.
     * @param pname The string representation for a location to be used in the
     *            given domain. Must not be <b>null</b> or empty.
     * @return An StpLocation appropriate for the given domain.
     * @throws WvcmException if the input is <b>null</b> or empty or an otherwise
     *             acceptable input specifies a domain different from that
     *             requested.
     * 
     */
    StpLocation stpLocation(Domain domain,
                            String pname) throws WvcmException;

    /**
     * Constructs an StpLocation from the individual fields of a location 
     * specification that follows the stable selector scheme.
     * 
     * @param domain The domain of the resource.
     * @param rType The resource type segment of the name.
     * @param objectId The object id of the resource.
     * @param repoId The repository id of the resource.
     * 
     * @return A stable-selector scheme StpLocation object with the specified
     *         fields
     * 
     * @throws WvcmException If the location is not well-formed.
     */
    StpLocation stableSelector(StpProvider.Domain domain,
                               String rType,
                               String objectId,
                               String repoId) throws WvcmException;

    /**
     * Constructs a path-scheme location for a specified scheme and
     * domain.
     * @param  domain  The domain type of the repository
     * @param  namespace the particular path-scheme for the location
     * @param  path  The pathname of the resource. Each type of scheme has its
     * own encoding requirements.
     *
     * @return  An StpLocation object for the specified repository
     *
     * @throws  WvcmException  If the selector is ill-formed
     */
    StpLocation pathLocation(StpProvider.Domain domain,
                             StpLocation.Namespace namespace,
                             String path) throws WvcmException;

    /**
     * Constructs a PNAME StpLocation for a pathname to a file system
     * object.
     * 
     * <p>
     * Shorthand for
     * </p>
     * 
     * <pre>
     * stpLocation(null, Namespace.PNAME, encodeFile(file.getPath()), null).
     * </pre>
     * 
     * @param file Abstract pathname. The file system object need not exist.
     * 
     * @return An StpLocation object with the specified fields
     * 
     * @throws WvcmException Thrown if an error occurs.
     */
    StpLocation filePathLocation(File file) throws WvcmException;

    /**
     * Constructs a PNAME Location for a pathname to a file system
     * object of a specific domain.
     * 
     * <p>
     * Shorthand for
     * </p>
     * 
     * <pre>
     * stpLocation(domain, Namespace.PNAME, encodeFile(file.getPath()), null).
     * </pre>
     * 
     * @param domain The StpProvider.Domain of the resource referenced by the
     *            given file object.
     * @param file Abstract pathname. The file system object need not exist.
     * 
     * @return An StpLocation object with the specified fields
     * 
     * @throws WvcmException Thrown if an error occurs.
     */
    StpLocation filePathLocation(StpProvider.Domain domain,
                                 File file) throws WvcmException;

    /**
     * Encodes a plain text string to be a single segment of the object name
     * field of a selector by escaping the segment delimiters, repository field
     * separator, and the escape character within the string.
     * 
     * <p>
     * Note 1: This method is provided as a convenience to clients. It is not
     * used by the implementation of the StpLocation class, which does not
     * change the encoding of segments unless explicitly stated in method
     * documentation.
     * 
     * @param str The String containing the plain text of the segment to be
     *            encoded.
     * 
     * @return The encoded form of the input String that is suitable for use as
     *         a segment of the name field within an object selector.
     */
    String encodeSegment(String str);

    /**
     * Undoes the character encoding performed by the {@link #encodeSegment
     * encodeSegment} method. This decoding does not need to be performed
     * segment by segment as it is only replacing the escape character from each
     * escaped character digraph.
     * 
     * @param str The segment of the selector to be decoded.
     * 
     * @return The decoded (plain text) segment.
     */
    String decodeSegment(String str);

    /**
     * Construct an StpAccessControlEntry value
     * 
     * @param principal The group, user, or role to which access is to be
     *            granted.
     * @param accessRights The access rights to be granted; must not be <b>null</b>
     *            or empty.
     * @return an instance of StpAccessControlEntry
     */
    StpAccessControlEntry buildAccessControlEntry(StpPrincipal principal,
                                                  AccessRight... accessRights);

    /**
     * Constructs, logs, and throws an StpException.
     * 
     * @param reasonCode The StpException.StpReasonCode to be associated with
     *            the exception. Must not be <b>null</b>.
     * @param messageData A structured message object or the data needed to
     *            construct one. The message data is passed as an Object[]{
     *            message-key, {arg1,...,argN}} if the message format is
     *            parameterized or as a String containing the message key if it
     *            is not. If the message-key is not found in the resource bundle
     *            it is treated as the format string. Must not be <b>null</b>.
     * @param resource The Resource argument to WvcmException
     * @param nestedExceptions The Throwable[] argument to WvcmException
     */
    void raise(StpReasonCode reasonCode,
               Object messageData,
               Resource resource,
               Throwable... nestedExceptions) throws WvcmException;

    /**
     * Constructs, logs, and throws an StpPropertyException.
     * 
     * @param stpReasonCode The StpReasoCode that defines this exception.
     * @param messageData A structured message object or the data needed to
     *            construct one. The message data is passed as an Object[]{
     *            message-key, {arg1,...,argN}} if the message format is
     *            parameterized or as a String containing the message key if it
     *            is not. If the message-key is not found in the resource bundle
     *            it is treated as the format string. Must not be <b>null</b>.
     * @param resource The resource causing this exception.
     * @param nestedExceptions Subordinate Exceptions that lead to the throwing
     *            of this StpException.
     * @param propertyName The name of the property with which this exception is
     *            associated.
     */
    public void raise(StpReasonCode stpReasonCode,
                      Object messageData,
                      Resource resource,
                      PropertyName propertyName,
                      Throwable... nestedExceptions) throws WvcmException;

    /**
     * Constructs, logs, and throws an StpPartialResultsException.
     * 
     * <p>
     * The reason code is always StpReasonCode.PARTIAL_RESULTS.
     * 
     * @param messageData A structured message object or the data needed to
     *            construct one. The message data is passed as an Object[]{
     *            message-key, {arg1,...,argN}} if the message format is
     *            parameterized or as a String containing the message key if it
     *            is not. If the message-key is not found in the resource bundle
     *            it is treated as the format string. Must not be <b>null</b>.
     * @param resource The resource that caused the exception.
     * @param okList A list of resources that were successfully retrieved
     *            and processed. Can be <b>null</b> or empty.
     * @param enumerationError <b>true</b> if accessing some of the resources
     *            named as targets was not attempted.
     * @param name The PropertyName of the property with which this exception is
     *            associated. May be <b>null</b> if the exception is not associated
     *            with a property.
     * @param nestedExceptions A list of exceptions each of which identifies a
     *            resource that caused an error when accessed.
     */
    public void raise(Object messageData,
                      Resource resource,
                      ResourceList<? extends Resource> okList,
                      boolean enumerationError,
                      PropertyName name,
                      Throwable... nestedExceptions) throws WvcmException;
}
